<?php

/**
* WPV_Cache
*
* Caching class for Views.
*
* This class is useful on a series of scenarios:
* 	- Parametric search with dependencies or counters, including post relationship filters.
* 	- Etc.
*
* @since 1.12
*/

class WPV_Cache {
	
	static $stored_cache = array();
	static $stored_cache_extended_for_post_relationship = array();
	static $stored_relationship_cache = array();
	
	function __construct() {
		
	}
	
	/**
	* generate_native_cache
	*
	* Mimics the caching construction of WordPress so we can use it for counting posts.
	* update_postmeta_cache should get cached data, so we avoind further queries for postmeta.
	* We still need to generate the cache for the given taxonomies.
	*
	* @param $id_posts array of post IDs
	* @param $f_data array of data to pseudo-cache
	*    'tax' => array of taxonomy names to cache
	*
	* @return (array) cached data compatible with the native $wp_object_cache->cache format
	*
	* @uses update_postmeta_cache
	*
	* @since 1.12
	*/

	static function generate_native_cache( $id_posts = array(), $f_data = array() ) {
		$f_taxes = ( isset( $f_data['tax'] ) ) ? $f_data['tax'] : array();
		$cache_post_meta = array();
		// Sanitize $id_posts
		// It usually comes from a WP_Query, but still
		$id_posts = array_map( 'esc_attr', $id_posts );
		$id_posts = array_map( 'trim', $id_posts );
		// is_numeric does sanitization
		$id_posts = array_filter( $id_posts, 'is_numeric' );
		$id_posts = array_map( 'intval', $id_posts );
		
		$cache_post_meta = update_postmeta_cache( $id_posts );
		
		//Then, the taxonomies
		// Note that the settings might be polluted by non-existing taxonomies
		$cache_post_taxes = array();
		$current_taxonomies = get_taxonomies( '', 'names' );
		$f_taxes = array_intersect( $f_taxes, $current_taxonomies );
		$f_taxes = array_values( $f_taxes );
		if ( 
			! empty( $f_taxes ) 
			&& ! empty( $id_posts ) 
		) {
			$terms = wp_get_object_terms( $id_posts, $f_taxes, array('fields' => 'all_with_object_id') );
			if ( is_wp_error( $terms ) ) {
				$terms = array();
			}
			$object_terms = array();
			foreach ( (array) $terms as $term ) {
				$object_terms[ $term->object_id ][ $term->taxonomy ][ $term->term_id ] = $term;
			}
			foreach ( $id_posts as $id_needed ) {
				foreach ( $f_taxes as $taxonomy ) {
					if ( ! isset( $object_terms[ $id_needed ][ $taxonomy ] ) ) {
						if ( ! isset( $object_terms[ $id_needed ] ) ) {
							$object_terms[ $id_needed ] = array();
						}
						$object_terms[ $id_needed ][ $taxonomy ] = array();
					}
				}
			}
			foreach ( $object_terms as $post_id => $value ) {
				foreach ( $value as $taxonomy => $terms ) {
					if ( ! isset( $cache_post_taxes[ $taxonomy . '_relationships' ] ) ) {
						$cache_post_taxes[ $taxonomy . '_relationships' ] = array();
					}
					$cache_post_taxes[ $taxonomy . '_relationships' ][ $post_id ] = $terms;
				}
			}
		}
		
		$cache_combined = array();
		$cache_combined['post_meta'] = $cache_post_meta;
		foreach ( $cache_post_taxes as $tax_key => $tax_cached_values ) {
			$cache_combined[ $tax_key ] = $tax_cached_values;
		}
		
		self::$stored_cache = $cache_combined;
		
		return $cache_combined;
	}
	
	/**
	* generate_auxiliar_cache
	*
	* Mimics the caching construction of WordPress so we can use it for counting posts.
	* Caches data for the passed custom fields and taxonomies, without adding it to self::$stored_cache
	*
	* @param $id_posts array of post IDs
	* @param $f_data array of data to pseudo-cache
	*    'tax' => array of taxonomy names to cache
	*    'cf'  => array of field meta_key's to cache
	*
	* @return (array) cached data compatible with the native $wp_object_cache->cache format
	*
	* @since 1.12
	*/

	static function generate_auxiliar_cache( $id_posts = array(), $f_data = array() ) {
		$cache_combined = self::$stored_cache;
		$f_fields = ( isset( $f_data['cf'] ) ) ? $f_data['cf'] : array();
		$f_taxes = ( isset( $f_data['tax'] ) ) ? $f_data['tax'] : array();
		// Sanitize $id_posts
		// It usually comes from a WP_Query, but still
		$id_posts = array_map( 'esc_attr', $id_posts );
		$id_posts = array_map( 'trim', $id_posts );
		// is_numeric does sanitization
		$id_posts = array_filter( $id_posts, 'is_numeric' );
		$id_posts = array_map( 'intval', $id_posts );
		$id_posts_postmeta_matched = array();
		$id_posts_postmeta_missed = array();
		// Clear $id_posts from posts already cached
		// Also, make sure $cache_combined['post_meta'] is an array
		if ( 
			isset( $cache_combined['post_meta'] )
			&& is_array( $cache_combined['post_meta'] )
		) {
			$exclude_ids = array_keys( $cache_combined['post_meta'] );
			$id_posts = array_diff( $id_posts, $exclude_ids );
		} else {
			$cache_combined['post_meta'] = array();
		}
		// First, the post_meta
		if ( 
			! empty( $f_fields ) 
			&& ! empty( $id_posts ) 
		) {
			global $wpdb;
			$id_list = implode( ',', $id_posts );
			$f_fields_count = count( $f_fields );
			$f_fields_placeholders = array_fill( 0, $f_fields_count, '%s' );
			$meta_list = $wpdb->get_results( 
				$wpdb->prepare(
					"SELECT post_id, meta_key, meta_value 
					FROM {$wpdb->postmeta} 
					WHERE post_id IN ({$id_list}) 
					AND meta_key IN (" . implode( ",", $f_fields_placeholders ) . ") 
					ORDER BY post_id ASC", 
					$f_fields
				),
				ARRAY_A 
			);
			if ( ! empty( $meta_list ) ) {
				foreach ( $meta_list as $metarow) {
					$mpid = intval( $metarow['post_id'] );
					$mkey = $metarow['meta_key'];
					$mval = $metarow['meta_value'];
					if ( ! in_array( $mpid, $id_posts_postmeta_matched ) ) {
						$id_posts_postmeta_matched[] = $mpid;
					}
					if (
						isset( $cache_combined['post_meta'][ $mpid ] )
					) {
						// The post has already been cached, let's check whether its meta key has been cached too
						if ( ! isset( $cache_combined['post_meta'][ $mpid ][ $mkey ] ) ) {
							$cache_combined['post_meta'][ $mpid ][ $mkey ] = array();
							$cache_combined['post_meta'][ $mpid ][ $mkey ][] = $mval;
						}
					} else {
						// We add to the $cache_combined['post_meta']
						$cache_combined['post_meta'][ $mpid ] = array();
						$cache_combined['post_meta'][ $mpid ][ $mkey ] = array();
						$cache_combined['post_meta'][ $mpid ][ $mkey ][] = $mval;
					}
				}
			}
			// Fill the gaps
			$id_posts_postmeta_missed = array_diff( $id_posts, $id_posts_postmeta_matched );
			foreach ( $id_posts_postmeta_missed as $id_needed ) {
				if ( ! isset( $cache_combined['post_meta'][ $id_needed ] ) ) {
					$cache_combined['post_meta'][ $id_needed ] = array();
				}
			}
		}
		
		//Then, the taxonomies
		$cache_post_taxes = array();
		$current_taxonomies = get_taxonomies( '', 'names' );
		$f_taxes = array_intersect( $f_taxes, $current_taxonomies );
		$f_taxes = array_values( $f_taxes );
		if ( 
			! empty( $f_taxes ) 
			&& ! empty( $id_posts ) 
		) {
			$terms = wp_get_object_terms( $id_posts, $f_taxes, array('fields' => 'all_with_object_id') );
			if ( is_wp_error( $terms ) ) {
				$terms = array();
			}
			$object_terms = array();
			foreach ( (array) $terms as $term ) {
				$object_terms[ $term->object_id ][ $term->taxonomy ][ $term->term_id ] = $term;
			}
			foreach ( $id_posts as $id_needed ) {
				foreach ( $f_taxes as $taxonomy ) {
					if ( ! isset( $object_terms[ $id_needed ][ $taxonomy ] ) ) {
						if ( ! isset( $object_terms[ $id_needed ] ) ) {
							$object_terms[ $id_needed ] = array();
						}
						$object_terms[ $id_needed ][ $taxonomy ] = array();
					}
				}
			}
			foreach ( $object_terms as $post_id => $value ) {
				foreach ( $value as $taxonomy => $terms ) {
					if ( ! isset( $cache_post_taxes[ $taxonomy . '_relationships' ] ) ) {
						$cache_post_taxes[ $taxonomy . '_relationships' ] = array();
					}
					$cache_post_taxes[ $taxonomy . '_relationships' ][ $post_id ] = $terms;
				}
			}
		}
		foreach ( $cache_post_taxes as $tax_key => $tax_cached_values ) {
			if ( isset( $cache_combined[ $tax_key ] ) ) {
				$cache_combined[ $tax_key ] = array_merge( $cache_combined[ $tax_key ], $tax_cached_values );
			} else {
				$cache_combined[ $tax_key ] = $tax_cached_values;
			}
		}
			
		return $cache_combined;
	}
	
	/**
	* generate_cache
	*
	* Mimics the caching construction of WordPress so we can use it for counting posts.
	* Caches data for the passed custom fields and taxonomies, and adds it to self::$stored_cache
	*
	* @param $id_posts array of post IDs
	* @param $f_data array of data to pseudo-cache
	*    'tax' => array of taxonomy names to cache
	*    'cf'  => array of field meta_key's to cache
	*
	* @uses self::generate_auxiliar_cache
	*
	* @return (array) cached data compatible with the native $wp_object_cache->cache format
	*
	* @since 1.12
	*/

	static function generate_cache( $id_posts = array(), $f_data = array() ) {
		$cache_combined = self::generate_auxiliar_cache( $id_posts, $f_data );
		self::$stored_cache = $cache_combined;
		return $cache_combined;
	}
	
	/**
	* generate_cache_extended_for_post_relationship
	*
	* Mimics the caching construction of WordPress so we can use it for counting posts.
	* Caches data for the passed custom fields and taxonomies, and adds it to self::$stored_cache_extended_for_post_relationship
	* Used when rendering post relationship filters with dependency or counters,
	* as we need to generate a specific query that avoinds the filter by the current post type in the relationship ree, if it exists
	*
	* @param $id_posts array of post IDs
	* @param $f_data array of data to pseudo-cache
	*    'tax' => array of taxonomy names to cache
	*    'cf'  => array of field meta_key's to cache
	*
	* @uses self::generate_auxiliar_cache
	*
	* @return (array) cached data compatible with the native $wp_object_cache->cache format
	*
	* @since 1.12
	*/
	
	static function generate_cache_extended_for_post_relationship( $id_posts = array(), $f_data = array() ) {
		$cache_combined = self::generate_auxiliar_cache( $id_posts, $f_data );
		self::$stored_cache_extended_for_post_relationship = $cache_combined;
		return $cache_combined;
	}
	
	/**
	* generate_post_relationship_tree_cache
	*
	* Generates data for counters and disable/hide elements in a post relationship parametric filter
	*
	* @param $tree		string	greater-than separated list of ancestors, in a top-to-bottom order
	* @param $count		bool	whether the count should return the number of matches or just a true/false statement
	*
	* @return
	*
	* @since 1.12
	*/
	
	static function generate_post_relationship_tree_cache( $tree, $count = true ) {
		$tree_array = explode( '>', $tree );
		$tree_real = array_reverse( $tree_array );
		$tree_remove = array_shift( $tree_real );
		$tree_ground = end( $tree_array );
		$tree_roof = reset( $tree_array );
		$current_post_ids = array();
		$counters = array();
		global $wpdb;
		$cache_combined = self::$stored_cache_extended_for_post_relationship;
		if ( 
			isset( $cache_combined['post_meta'] )
			&& is_array( $cache_combined['post_meta'] )
		) {
			$cached_postmeta = $cache_combined['post_meta'];
			$field = '_wpcf_belongs_' . $tree_ground . '_id';
			// We can not use wp_list_pluck here because it produces warnings when ( ! isset( $value[ $field ] ) )
			foreach ( $cached_postmeta as $key => $value ) {
				if ( isset( $value[ $field ] ) ) {
					$cached_postmeta[ $key ] = $value[ $field ];
				} else {
					unset( $cached_postmeta[ $key ] );
				}
			}
			$current_post_ids = call_user_func_array('array_merge', $cached_postmeta );
			foreach ( $current_post_ids as $cpi ) {
				$meta_criteria_to_filter = array( '_wpcf_belongs_' . $tree_ground . '_id' => array( $cpi ) );
				$data = array();
				$data['list'] = $cache_combined['post_meta'];
				$data['args'] = $meta_criteria_to_filter;
				$data['kind'] = '';
				$data['comparator'] = 'equal';
				$data['count_matches'] = $count;
				$counters[ $cpi ] = array(
					'type'	=> $tree_ground,
					'count'	=> wpv_list_filter_checker( $data )
				);
			}
		}
		foreach ( $tree_real as $tree_branch ) {
			$current_post_ids = array_map( 'esc_attr', $current_post_ids );
			$current_post_ids = array_map( 'trim', $current_post_ids );
			$current_post_ids = array_filter( $current_post_ids, 'is_numeric' );
			$current_post_ids = array_map( 'intval', $current_post_ids );
			if ( count( $current_post_ids ) > 0 ) {
				$future_post_ids = $wpdb->get_results(
					$wpdb->prepare(
						"SELECT post_id, meta_value
						FROM {$wpdb->postmeta}
						WHERE meta_key = %s 
						AND post_id IN ('" . implode( "','", $current_post_ids ) . "')",
						'_wpcf_belongs_' . $tree_branch . '_id'
					),
					ARRAY_A
				);
				foreach ( $future_post_ids as $fpid ) {
					$add = isset( $counters[ $fpid['post_id'] ] ) ? $counters[ $fpid['post_id'] ]['count'] : 1;
					if ( ! isset( $counters[ $fpid['meta_value'] ] ) ) {
						$counters[ $fpid['meta_value'] ] = array(
							'type'	=> $tree_branch,
							'count'	=> 0
						);
					}
					$counters[ $fpid['meta_value'] ]['count'] = isset( $counters[ $fpid['meta_value'] ] ) ? ( $counters[ $fpid['meta_value'] ]['count'] + $add ) : $add;
				}
				$current_post_ids = wp_list_pluck( $future_post_ids, 'meta_value' );
			} else {
				$current_post_ids = array();
			}
		}
		self::$stored_relationship_cache = $counters;
		return $counters;
	}
	
}

$WPV_Cache = new WPV_Cache();